@@ -14,6 +14,7 @@ import logging
 import os
 import re
 import sys
+import numpy as np
 
 import mako.lookup
 import mako.template
@@ -32,6 +33,9 @@ RFNOC_CORE_DIR = os.path.join('rfnoc', 'core')
 # Path to the system's bash executable
 BASH_EXECUTABLE = '/bin/bash' # FIXME this should come from somewhere
 
+# Supported protocol version
+RFNOC_PROTO_VERSION = "1.0"
+
 # Map device names to the corresponding directory under usrp3/top
 DEVICE_DIR_MAP = {
     'x300': 'x300',
@@ -44,6 +48,7 @@ DEVICE_DIR_MAP = {
     'n320': 'n3xx',
     'x400': 'x400',
     'x410': 'x400',
+    'x440': 'x400',
 }
 
 # Picks the default make target per device
@@ -216,9 +221,9 @@ class ImageBuilderConfig:
         self.__dict__.update(**config)
         self.blocks = blocks
         self.device = device
+        self._update_sep_defaults()
         self._check_configuration()
         self._check_deprecated_signatures()
-        self._update_sep_defaults()
         self._set_indices()
         self._collect_noc_ports()
         self._collect_io_ports()
@@ -262,6 +267,71 @@ class ImageBuilderConfig:
         failure = None
         if not any([bool(sep["ctrl"]) for sep in self.stream_endpoints.values()]):
             failure = "At least one streaming endpoint needs to have ctrl enabled"
+        # Check RFNoC protocol version. Use latest if it was not specified.
+        requested_version = (
+            self.rfnoc_version
+            if hasattr(self, "rfnoc_version")
+            else RFNOC_PROTO_VERSION
+        )
+        [requsted_major, requested_minor, *_] = requested_version.split('.')
+        [supported_major, supported_minor, *_] = RFNOC_PROTO_VERSION.split('.')
+        if requsted_major != supported_major or requested_minor > supported_minor:
+            failure = (
+                "Requested RFNoC protocol version (rfnoc_version) " +
+                requested_version +
+                " is ahead of latest version "
+                + RFNOC_PROTO_VERSION
+            )
+        elif requested_version != RFNOC_PROTO_VERSION:
+            logging.warning(
+                "Generating code for latest RFNoC protocol version "
+                + RFNOC_PROTO_VERSION
+                + " instead of requested "
+                + requested_version
+            )
+        self.rfnoc_version = RFNOC_PROTO_VERSION
+        # Give block_chdr_width a default value
+        if not hasattr(self, "block_chdr_width"):
+            self.block_chdr_width = self.chdr_width
+        # Check crossbar_routes
+        num_tas = len(self.device.transports)
+        num_seps = len(self.stream_endpoints)
+        num_ports = num_tas + num_seps
+        if hasattr(self, "crossbar_routes"):
+            # Check that crossbar_routes is an NxN array of ones and zeros
+            routes = np.array(self.crossbar_routes)
+            num_rows = routes.shape[0]
+            num_cols = routes.shape[1]
+            num_digits = np.count_nonzero(routes == 0) + np.count_nonzero(routes == 1)
+            if routes.shape != (num_ports, num_ports) or num_digits != num_ports**2:
+                failure = (
+                    "crossbar_routes must be a "
+                    + str(num_ports)
+                    + " by "
+                    + str(num_ports)
+                    + " binary array"
+                )
+        else:
+            # Give crossbar_routes a default value
+            routes = np.ones([num_ports, num_ports])
+            # Disable all TA to TA paths, except loopback (required for discovery)
+            for i in range(0, num_tas):
+                routes[i, 0:num_tas] = 0
+                routes[i, i] = 1
+            self.crossbar_routes = routes.tolist()
+        # Check SEP has buff_size and buff_size_bytes parameters
+        for sep_name in self.stream_endpoints:
+            sep = self.stream_endpoints[sep_name]
+            if "buff_size" not in sep and "buff_size_bytes" not in sep:
+                failure = (
+                    f"You must specify buff_size or buff_size_bytes for {sep_name}"
+                )
+            # Initialize the one not set by user (schema doesn't allow both)
+            if "buff_size_bytes" in sep:
+                sep["buff_size"] = sep["buff_size_bytes"] // (sep["chdr_width"] // 8)
+            elif "buff_size" in sep:
+                sep["buff_size_bytes"] = sep["buff_size"] * (sep["chdr_width"] // 8)
+        # Handle any errors
         if failure:
             logging.error(failure)
             raise ValueError(failure)
@@ -275,6 +345,8 @@ class ImageBuilderConfig:
                 self.stream_endpoints[sep]["num_data_i"] = 1
             if "num_data_o" not in self.stream_endpoints[sep]:
                 self.stream_endpoints[sep]["num_data_o"] = 1
+            if "chdr_width" not in self.stream_endpoints[sep]:
+                self.stream_endpoints[sep]["chdr_width"] = self.chdr_width
 
     def _set_indices(self):
         """
@@ -759,7 +831,7 @@ def write_edges(config, destination):
                           ((dst[0] << 6) | dst[1])))
 
 
-def write_verilog(config, destination, source, source_hash):
+def write_verilog(config, destination, args):
     """
     Generates rfnoc_image_core.v file for the device.
 
@@ -770,8 +842,7 @@ def write_verilog(config, destination, source, source_hash):
     template engine.
     :param config: ImageBuilderConfig derived from script parameter
     :param destination: Filepath to write to
-    :param source: Filepath to the image YAML/GRC to generate from
-    :param source_hash: Source file hash value
+    :param args: Dictionary of arguments for the code generation
     :return: None
     """
     template_dir = os.path.join(os.path.dirname(__file__), "templates")
@@ -785,8 +856,7 @@ def write_verilog(config, destination, source, source_hash):
     try:
         block = tpl.render(**{
             "config": config,
-            "source": source,
-            "source_hash": source_hash,
+            "args": args,
             })
     except:
         print(exceptions.text_error_template().render())
@@ -797,13 +867,12 @@ def write_verilog(config, destination, source, source_hash):
         image_core_file.write(block)
 
 
-def write_verilog_header(config, destination, source, source_hash):
+def write_verilog_header(config, destination, args):
     """
     Generates rfnoc_image_core.vh file for the device.
     :param config: ImageBuilderConfig derived from script parameter
     :param destination: Filepath to write to
-    :param source: Filepath to the image YAML/GRC to generate from
-    :param source_hash: Source file hash value
+    :param args: Dictionary of arguments for the code generation
     :return: None
     """
     template_dir = os.path.join(os.path.dirname(__file__), "templates")
@@ -817,8 +886,7 @@ def write_verilog_header(config, destination, source, source_hash):
     try:
         block = tpl.render(**{
             "config": config,
-            "source": source,
-            "source_hash": source_hash,
+            "args": args,
             })
     except:
         print(exceptions.text_error_template().render())
@@ -1015,13 +1083,13 @@ def build_image(config, fpga_path, config_path, device, **args):
     write_verilog(
         builder_conf,
         image_core_path,
-        source=args.get('source'),
-        source_hash=args.get('source_hash'))
+        args,
+    )
     write_verilog_header(
         builder_conf,
         image_core_header_path,
-        source=args.get('source'),
-        source_hash=args.get('source_hash'))
+        args,
+    )
     write_build_env()
 
     if "generate_only" in args and args["generate_only"]:
