@@ -14,7 +14,7 @@ from os import path
 from collections import namedtuple
 from usrp_mpm import lib # Pulls in everything from C++-land
 from usrp_mpm import tlv_eeprom
-from usrp_mpm.cores import WhiteRabbitRegsControl
+from usrp_mpm.compat_num import CompatNumber
 from usrp_mpm.components import ZynqComponents
 from usrp_mpm.sys_utils import dtoverlay
 from usrp_mpm.sys_utils import ectool
@@ -30,21 +30,18 @@ from usrp_mpm.periph_manager.x4xx_periphs import CtrlportRegs
 from usrp_mpm.periph_manager.x4xx_dio_control import DioControl
 from usrp_mpm.periph_manager.x4xx_periphs import QSFPModule
 from usrp_mpm.periph_manager.x4xx_periphs import get_temp_sensor
-from usrp_mpm.periph_manager.x4xx_mb_cpld import MboardCPLD
+from usrp_mpm.periph_manager.x4xx_mb_cpld import make_mb_cpld_ctrl
 from usrp_mpm.periph_manager.x4xx_clk_aux import ClockingAuxBrdControl
-from usrp_mpm.periph_manager.x4xx_clk_mgr import X4xxClockMgr
+from usrp_mpm.periph_manager.x4xx_clock_mgr import X4xxClockManager
 from usrp_mpm.periph_manager.x4xx_gps_mgr import X4xxGPSMgr
 from usrp_mpm.periph_manager.x4xx_rfdc_ctrl import X4xxRfdcCtrl
+from usrp_mpm.periph_manager.x4xx_clock_policy import get_clock_policy
 from usrp_mpm.dboard_manager.x4xx_db_iface import X4xxDboardIface
-from usrp_mpm.dboard_manager.zbx import ZBX
 
-
-X400_DEFAULT_EXT_CLOCK_FREQ = 10e6
-X400_DEFAULT_MASTER_CLOCK_RATE = 122.88e6
-X400_DEFAULT_TIME_SOURCE = X4xxClockMgr.TIME_SOURCE_INTERNAL
-X400_DEFAULT_CLOCK_SOURCE = X4xxClockMgr.CLOCK_SOURCE_INTERNAL
+X400_FPGA_COMPAT = (8, 0)
+# The compat number at which remote streaming was added:
+X400_REMOTE_STREAMING_COMPAT = (7, 9)
 X400_DEFAULT_ENABLE_PPS_EXPORT = True
-X400_FPGA_COMPAT = (7, 7)
 X400_DEFAULT_TRIG_DIRECTION = ClockingAuxBrdControl.DIRECTION_OUTPUT
 X400_MONITOR_THREAD_INTERVAL = 1.0 # seconds
 QSFPModuleConfig = namedtuple("QSFPModuleConfig", "modprs modsel devsymbol")
@@ -151,7 +148,9 @@ class x4xx(ZynqComponents, PeriphManagerBase):
     # See PeriphManagerBase for documentation on these fields. We try and keep
     # them in the same order as they are in PeriphManagerBase for easier lookup.
     #########################################################################
-    pids = {0x0410: 'x410'}
+    pids = {
+        0x0410: 'x410',
+        0x0440: 'x440'}
     description = "X400-Series Device"
     eeprom_search = PeriphManagerBase._EepromSearch.SYMBOL
     # This is not in the overridables section from PeriphManagerBase, but we use
@@ -171,12 +170,12 @@ class x4xx(ZynqComponents, PeriphManagerBase):
         'fan0': 'get_fan0_sensor',
         'fan1': 'get_fan1_sensor',
         'temp_fpga' : 'get_fpga_temp_sensor',
-        'temp_internal' : 'get_internal_temp_sensor',
         'temp_main_power' : 'get_main_power_temp_sensor',
         'temp_scu_internal' : 'get_scu_internal_temp_sensor',
     }
     db_iface = X4xxDboardIface
     dboard_eeprom_magic = eeprom_magic
+    # Note: Daughterboard classes also carry updateable_components information.
     updateable_components = {
         'fpga': {
             'callback': "update_fpga",
@@ -186,7 +185,7 @@ class x4xx(ZynqComponents, PeriphManagerBase):
             'compatibility': {
                 'fpga': {
                     'current': X400_FPGA_COMPAT,
-                    'oldest': (7, 0),
+                    'oldest': (8, 0),
                 },
                 'cpld_ifc' : {
                     'current': (2, 0),
@@ -196,14 +195,6 @@ class x4xx(ZynqComponents, PeriphManagerBase):
                     'current': (1, 0),
                     'oldest': (1, 0),
                 },
-                'rf_core_100m': {
-                    'current': (1, 0),
-                    'oldest': (1, 0),
-                },
-                'rf_core_400m': {
-                    'current': (1, 0),
-                    'oldest': (1, 0),
-                },
             }
         },
         'dts': {
@@ -218,15 +209,10 @@ class x4xx(ZynqComponents, PeriphManagerBase):
     # End of overridables from PeriphManagerBase
     ###########################################################################
 
-
     # X400-specific settings
     # Label for the mboard UIO
     mboard_regs_label = "mboard-regs"
     ctrlport_regs_label = "ctrlport-mboard-regs"
-    # Label for the white rabbit UIO
-    wr_regs_label = "wr-regs"
-    # Override the list of updateable components
-    # X4xx specific discoverable features
 
     @classmethod
     def generate_device_info(cls, eeprom_md, mboard_info, dboard_infos):
@@ -286,13 +272,8 @@ class x4xx(ZynqComponents, PeriphManagerBase):
         self._tear_down = False
         self._rpu_initialized = False
         self._status_monitor_thread = None
-        self._master_clock_rate = None
         self._gps_mgr = None
-        self._clk_mgr = None
-        self._safe_sync_source = {
-            'clock_source': X400_DEFAULT_CLOCK_SOURCE,
-            'time_source': X400_DEFAULT_TIME_SOURCE,
-        }
+        self.clk_mgr = None
         self.rfdc = None
         self.mboard_regs_control = None
         self.ctrlport_regs = None
@@ -304,7 +285,7 @@ class x4xx(ZynqComponents, PeriphManagerBase):
             # We need to init dio_control separately from peripherals
             # since it needs information about available dboards
             self._init_dio_control(args)
-            self._clk_mgr.set_dboard_reset_cb(
+            self.clk_mgr.set_dboard_reset_cb(
                 lambda enable: [db.reset_clock(enable) for db in self.dboards])
         except Exception as ex:
             self.log.error("Failed to initialize motherboard: %s", str(ex), exc_info=ex)
@@ -315,13 +296,14 @@ class x4xx(ZynqComponents, PeriphManagerBase):
             return
         try:
             if not args.get('skip_boot_init', False):
+                args['boot_init'] = True
                 self.init(args)
         except Exception as ex:
             self.log.warning("Failed to initialize device on boot: %s", str(ex))
 
         # Freeze the RFDC calibration by default
-        self.rfdc.set_cal_frozen(1, 1, "both")
-        self.rfdc.set_cal_frozen(1, 0, "both")
+        self.rfdc.set_cal_frozen(1, 1, "all")
+        self.rfdc.set_cal_frozen(1, 0, "all")
 
     # The parent class versions of these functions require access to self, but
     # these versions don't.
@@ -342,7 +324,14 @@ class x4xx(ZynqComponents, PeriphManagerBase):
     # pylint: enable=no-self-use
 
     def _check_fpga_compat(self):
-        " Throw an exception if the compat numbers don't match up "
+        """
+        Read the FPGA compat number and throw an exception if the compat numbers
+        don't match up (i.e., this version of MPM requires a newer FPGA compat
+        number).
+
+        If the numbers match, we check the FPGA compat number for available
+        features.
+        """
         actual_compat = self.mboard_regs_control.get_compat_number()
         self.log.debug("Actual FPGA compat number: {:d}.{:d}".format(
             actual_compat[0], actual_compat[1]
@@ -354,6 +343,11 @@ class x4xx(ZynqComponents, PeriphManagerBase):
             fail_on_old_minor=False,
             log=self.log
         )
+        if CompatNumber(actual_compat) >= CompatNumber(X400_REMOTE_STREAMING_COMPAT):
+            self.fpga_features.add('remote_udp_streaming')
+        self.log.debug(
+            "FPGA supports the following features: {}"
+            .format(", ".join(self.fpga_features)))
 
     def _init_gps_mgr(self):
         """
@@ -420,6 +414,13 @@ class x4xx(ZynqComponents, PeriphManagerBase):
         """
         Turn on all peripherals. This may throw an error on failure, so make
         sure to catch it.
+
+        Reminder: This is the first function that is run that actually talks to
+        hardware *after* PeriphManagerBase.__init__() has been executed. The
+        only bits of hardware that have been initialized/read out are the
+        motherboard and daughterboard EEPROMs.
+        After this, we initialize the daughterboards, so we have to make sure
+        everything else, in particular the clocks, are running as desired.
         """
         # Sanity checks
         assert self.mboard_info.get('product') in self.pids.values(), \
@@ -436,59 +437,42 @@ class x4xx(ZynqComponents, PeriphManagerBase):
             self._clocking_auxbrd = ClockingAuxBrdControl()
             self.log.trace("Initialized Clocking Aux Board controls")
             has_gps = self._clocking_auxbrd.is_gps_supported()
+            self._add_public_methods(self._clocking_auxbrd, "clkaux")
         except RuntimeError:
             self.log.warning(
                 "GPIO I2C bus could not be found for the Clocking Aux Board, "
                 "disabling Clocking Aux Board functionality.")
             self._clocking_auxbrd = None
-        self._safe_sync_source = {
-            'clock_source': X4xxClockMgr.CLOCK_SOURCE_MBOARD,
-            'time_source': X4xxClockMgr.TIME_SOURCE_INTERNAL,
-        }
-
-        initial_clock_source = args.get('clock_source', X400_DEFAULT_CLOCK_SOURCE)
-        if self._clocking_auxbrd:
-            self._add_public_methods(self._clocking_auxbrd, "clkaux")
-        else:
-            initial_clock_source = X4xxClockMgr.CLOCK_SOURCE_MBOARD
 
         # Init CPLD before talking to clocking ICs
         cpld_spi_node = dt_symbol_get_spidev('mb_cpld')
-        self.cpld_control = MboardCPLD(cpld_spi_node, self.log)
-        self.cpld_control.check_signature()
-        self.cpld_control.check_compat_version()
-        self.cpld_control.trace_git_hash()
+        # This factory function will check signature and compat-rev, and
+        # therefore could throw if the CPLD is not compatible.
+        self.cpld_control = make_mb_cpld_ctrl(cpld_spi_node, self.log)
 
         self._assert_rfdc_powered()
         # Init clocking after CPLD as the SPLL communication is relying on it.
         # We try and guess the correct master clock rate here based on defaults
-        # and args. Since we are still in __init__(), the args that come from mpm.conf
-        # are empty. We can't detect the real default MCR, because we need
-        # the RFDC controls for that -- but they won't work without clocks. So
-        # let's pick a sensible default MCR value, init the clocks, and fix the
-        # MCR value further down.
-        self._master_clock_rate = float(
-            args.get('master_clock_rate', X400_DEFAULT_MASTER_CLOCK_RATE))
-        sample_clock_freq, _, is_legacy_mode, _ = \
-            X4xxRfdcCtrl.master_to_sample_clk[self._master_clock_rate]
-        self._clk_mgr = X4xxClockMgr(
-            initial_clock_source,
-            time_source=args.get('time_source', X400_DEFAULT_TIME_SOURCE),
-            ref_clock_freq=float(args.get(
-                'ext_clock_freq', X400_DEFAULT_EXT_CLOCK_FREQ)),
-            sample_clock_freq=sample_clock_freq,
-            is_legacy_mode=is_legacy_mode,
+        # and args. Since we are still in __init__(), we don't know yet which
+        # MCR will be requested by the user.  We can't detect the real default
+        # MCR, because we need the RFDC controls for that -- but they won't
+        # work without clocks. So let's pick a sensible default MCR value, init
+        # the clocks, and fix the MCR value later (in init()).
+        self.clk_mgr = X4xxClockManager(
+            args,
+            clk_policy=get_clock_policy(self.mboard_info, self.dboard_infos, args, self.log),
             clk_aux_board=self._clocking_auxbrd,
             cpld_control=self.cpld_control,
             log=self.log)
         self._add_public_methods(
-            self._clk_mgr,
+            self.clk_mgr,
             prefix="",
             filter_cb=lambda name, method: not hasattr(method, '_norpc'),
             allow_overwrite=True
         )
 
-        # Overlay must be applied after clocks have been configured
+        # Overlay must be applied after clocks have been configured. This will
+        # load the FPGA bitfile.
         self.overlay_apply()
 
         # Init Mboard Regs
@@ -500,10 +484,11 @@ class x4xx(ZynqComponents, PeriphManagerBase):
         self.mboard_regs_control.set_serial_number(serial_number)
         self.mboard_regs_control.get_git_hash()
         self.mboard_regs_control.get_build_timestamp()
-        self._clk_mgr.mboard_regs_control = self.mboard_regs_control
+        # The clock manager needs access to this object -- we will make that
+        # available after the RFDC object is created.
 
         # Create control for RFDC
-        self.rfdc = X4xxRfdcCtrl(self._clk_mgr.get_spll_freq, self.log)
+        self.rfdc = X4xxRfdcCtrl(self.log)
         self._add_public_methods(
             self.rfdc, prefix="",
             filter_cb=lambda name, method: not hasattr(method, '_norpc')
@@ -511,18 +496,13 @@ class x4xx(ZynqComponents, PeriphManagerBase):
 
         self._update_fpga_type()
 
-        # Force reset the RFDC to ensure it is in a good state
-        self.rfdc.set_reset(reset=True)
-        self.rfdc.set_reset(reset=False)
-
-        # Synchronize SYSREF and clock distributed to all converters
-        self.rfdc.sync()
-        self._clk_mgr.set_rfdc_reset_cb(self.rfdc.set_reset)
-
-        # The initial default mcr only works if we have an FPGA with
-        # a decimation of 2. But we need the overlay applied before we
-        # can detect decimation, and that requires clocks to be initialized.
-        self.set_master_clock_rate(self.rfdc.get_default_mcr())
+        # Now that we have applied the overlay, we can query the FPGA
+        # capabilities. This will allow us to figure out a good default MCR,
+        # and we set that. Remember that earlier, when we turned on the clocks,
+        # we just picked any valid clock rate so we would be able to communicate
+        # with all the devices. After this call, all clocks will be set to useful
+        # and valid values.
+        self.clk_mgr.finalize_init(args, self.mboard_regs_control, self.rfdc)
 
         # Init ctrlport endpoint
         self.ctrlport_regs = CtrlportRegs(self.ctrlport_regs_label, self.log)
@@ -639,16 +619,9 @@ class x4xx(ZynqComponents, PeriphManagerBase):
         if self._clocking_auxbrd is not None:
             self._clocking_auxbrd.set_trig(False)
 
-        # If the caller has not specified clock_source or time_source, set them
-        # to the default values.
-        args['clock_source'] = args.get('clock_source', X400_DEFAULT_CLOCK_SOURCE)
-        args['time_source'] = args.get('time_source', X400_DEFAULT_TIME_SOURCE)
-        self.set_sync_source(args)
-
-        # If a Master Clock Rate was specified,
-        # re-configure the Sample PLL and all downstream clocks
-        if 'master_clock_rate' in args:
-            self.set_master_clock_rate(float(args['master_clock_rate']))
+        # Initialize all clocking-related stuff: Master clock rate, time/clock
+        # source, initialize all the PLLs, ...
+        self.clk_mgr.init(args)
 
         # Initialize CtrlportRegs (manually opens the UIO resource for faster access)
         self.ctrlport_regs.init()
@@ -663,9 +636,6 @@ class x4xx(ZynqComponents, PeriphManagerBase):
                 args.get('pps_export', X400_DEFAULT_ENABLE_PPS_EXPORT),
                 args.get('trig_direction', X400_DEFAULT_TRIG_DIRECTION)
                 )
-
-        for xport_mgr in self._xport_mgrs.values():
-            xport_mgr.init(args)
         return result
 
     def deinit(self):
@@ -676,17 +646,10 @@ class x4xx(ZynqComponents, PeriphManagerBase):
             self.log.warning(
                 "Cannot run deinit(), device was never fully initialized!")
             return
+        self.clk_mgr.deinit()
 
-        if self.get_ref_lock_sensor()['unit'] != 'locked':
-            self.log.error("ref clocks aren't locked, falling back to default")
-            source = {"clock_source": X400_DEFAULT_CLOCK_SOURCE,
-                      "time_source": X400_DEFAULT_TIME_SOURCE
-                     }
-            self.set_sync_source(source)
         super(x4xx, self).deinit()
         self.ctrlport_regs.deinit()
-        for xport_mgr in self._xport_mgrs.values():
-            xport_mgr.deinit()
 
     def tear_down(self):
         """
@@ -705,8 +668,10 @@ class x4xx(ZynqComponents, PeriphManagerBase):
         super(x4xx, self).tear_down()
         if self.dio_control is not None:
             self.dio_control.tear_down()
-        self.rfdc.tear_down()
-        self._clk_mgr.unset_cbs()
+        if self.rfdc:
+            self.rfdc.tear_down()
+        if self.clk_mgr:
+            self.clk_mgr.tear_down()
         # remove x4xx overlay
         active_overlays = self.list_active_overlays()
         self.log.trace("X4xx has active device tree overlays: {}".format(
@@ -715,39 +680,6 @@ class x4xx(ZynqComponents, PeriphManagerBase):
         for overlay in active_overlays:
             dtoverlay.rm_overlay(overlay)
 
-    ###########################################################################
-    # Transport API
-    ###########################################################################
-    # pylint: disable=no-self-use
-    def get_chdr_link_types(self):
-        """
-        This will only ever return a single item (udp).
-        """
-        return ["udp"]
-    # pylint: enable=no-self-use
-
-    def get_chdr_link_options(self, xport_type):
-        """
-        Returns a list of dictionaries. Every dictionary contains information
-        about one way to connect to this device in order to initiate CHDR
-        traffic.
-
-        The interpretation of the return value is very highly dependant on the
-        transport type (xport_type).
-        For UDP, the every entry of the list has the following keys:
-        - ipv4 (IP Address)
-        - port (UDP port)
-        - link_rate (bps of the link, e.g. 10e9 for 10GigE)
-        """
-        if xport_type not in self._xport_mgrs:
-            self.log.warning("Can't get link options for unknown link type: `{}'.")
-            return []
-        if xport_type == "udp":
-            return self._xport_mgrs[xport_type].get_chdr_link_options(
-                self.mboard_info['rpc_connection'])
-        # else:
-        return self._xport_mgrs[xport_type].get_chdr_link_options()
-
     ###########################################################################
     # Device info
     ###########################################################################
@@ -775,161 +707,9 @@ class x4xx(ZynqComponents, PeriphManagerBase):
         return db_gpio_version[0] > 0
 
     ###########################################################################
-    # Clock/Time API
+    # Clock/Time API. Note: Most of the API calls are defined in
+    # X4xxClockManager.
     ###########################################################################
-    def get_clock_sources(self):
-        """
-        Lists all available clock sources.
-        """
-        return self._clk_mgr.get_clock_sources()
-
-    def set_clock_source(self, *args):
-        """
-        Ensures the new reference clock source and current time source pairing
-        is valid and sets both by calling set_sync_source().
-        """
-        clock_source = args[0]
-        time_source = self._clk_mgr.get_time_source()
-        assert clock_source is not None
-        assert time_source is not None
-        if (clock_source, time_source) not in self._clk_mgr.valid_sync_sources:
-            old_time_source = time_source
-            if clock_source in (
-                    X4xxClockMgr.CLOCK_SOURCE_MBOARD,
-                    X4xxClockMgr.CLOCK_SOURCE_INTERNAL):
-                time_source = X4xxClockMgr.TIME_SOURCE_INTERNAL
-            elif clock_source == X4xxClockMgr.CLOCK_SOURCE_EXTERNAL:
-                time_source = X4xxClockMgr.TIME_SOURCE_EXTERNAL
-            elif clock_source == X4xxClockMgr.CLOCK_SOURCE_GPSDO:
-                time_source = X4xxClockMgr.TIME_SOURCE_GPSDO
-            self.log.warning(
-                f"Time source '{old_time_source}' is an invalid selection with "
-                f"clock source '{clock_source}'. "
-                f"Coercing time source to '{time_source}'")
-        self.set_sync_source({
-            "clock_source": clock_source, "time_source": time_source})
-
-    def set_clock_source_out(self, enable):
-        """
-        Allows routing the clock configured as source on the clk aux board to
-        the RefOut terminal. This only applies to internal, gpsdo and nsync.
-        """
-        self._clk_mgr.set_clock_source_out(enable)
-
-    def get_time_sources(self):
-        " Returns list of valid time sources "
-        return self._clk_mgr.get_time_sources()
-
-    def set_time_source(self, time_source):
-        """
-        Set a time source
-
-        This will call set_sync_source() internally, and use the current clock
-        source as a clock source. If the current clock source plus the requested
-        time source is not a valid combination, it will coerce the clock source
-        to a valid choice and print a warning.
-        """
-        clock_source = self._clk_mgr.get_clock_source()
-        assert clock_source is not None
-        assert time_source is not None
-        if (clock_source, time_source) not in self._clk_mgr.valid_sync_sources:
-            old_clock_source = clock_source
-            if time_source == X4xxClockMgr.TIME_SOURCE_QSFP0:
-                clock_source = X4xxClockMgr.CLOCK_SOURCE_MBOARD
-            elif time_source == X4xxClockMgr.TIME_SOURCE_INTERNAL:
-                clock_source = X4xxClockMgr.CLOCK_SOURCE_MBOARD
-            elif time_source == X4xxClockMgr.TIME_SOURCE_EXTERNAL:
-                clock_source = X4xxClockMgr.CLOCK_SOURCE_EXTERNAL
-            elif time_source == X4xxClockMgr.TIME_SOURCE_GPSDO:
-                clock_source = X4xxClockMgr.CLOCK_SOURCE_GPSDO
-            self.log.warning(
-                'Clock source {} is an invalid selection with time source {}. '
-                'Coercing clock source to {}'
-                .format(old_clock_source, time_source, clock_source))
-        self.set_sync_source(
-            {"time_source": time_source, "clock_source": clock_source})
-
-    def set_sync_source(self, args):
-        """
-        Selects reference clock and PPS sources. Unconditionally re-applies the
-        time source to ensure continuity between the reference clock and time
-        rates.
-        Note that if we change the source such that the time source is changed
-        to 'external', then we need to also disable exporting the reference
-        clock (RefOut and PPS-In are the same SMA connector).
-        """
-        # Check the clock source, time source, and combined pair are valid:
-        clock_source = args.get('clock_source', self._clk_mgr.get_clock_source())
-        if clock_source not in self.get_clock_sources():
-            raise ValueError(f'Clock source {clock_source} is not a valid selection')
-        time_source = args.get('time_source', self._clk_mgr.get_time_source())
-        if time_source not in self.get_time_sources():
-            raise ValueError(f'Time source {time_source} is not a valid selection')
-        if (clock_source, time_source) not in self._clk_mgr.valid_sync_sources:
-            raise ValueError(
-                f'Clock and time source pair ({clock_source}, {time_source}) is '
-                'not a valid selection')
-        # Sanity checks complete. Now check if we need to disable the RefOut.
-        # Reminder: RefOut and PPSIn share an SMA. Besides, you can't export an
-        # external clock. We are thus not checking for time_source == 'external'
-        # because that's a subset of clock_source == 'external'.
-        # We also disable clock exports for 'mboard', because the mboard clock
-        # does not get routed back to the clocking aux board and thus can't be
-        # exported either.
-        if clock_source in (X4xxClockMgr.CLOCK_SOURCE_EXTERNAL,
-                            X4xxClockMgr.CLOCK_SOURCE_MBOARD) and \
-                                    self._clocking_auxbrd:
-            self._clocking_auxbrd.export_clock(enable=False)
-        # Now the clock manager can do its thing.
-        ret_val = self._clk_mgr.set_sync_source(clock_source, time_source)
-        if ret_val == self._clk_mgr.SetSyncRetVal.NOP:
-            return
-        try:
-            # Re-set master clock rate. If this doesn't work, it will time out
-            # and throw an exception. We need to put the device back into a safe
-            # state in that case.
-            self.set_master_clock_rate(self._master_clock_rate)
-            # Restore the nco frequency to the same values as before the sync source
-            # was changed, to ensure the device transmission/acquisition continues at
-            # the requested frequency.
-            self.rfdc.rfdc_restore_nco_freq()
-            # Do the same for the calibration freeze state
-            self.rfdc.rfdc_restore_cal_freeze()
-        except RuntimeError as ex:
-            err = f"Setting clock_source={clock_source},time_source={time_source} " \
-                  f"failed, falling back to {self._safe_sync_source}. Error: " \
-                  f"{ex}"
-            self.log.error(err)
-            if args.get('__noretry__', False):
-                self.log.error("Giving up.")
-            else:
-                self.set_sync_source({**self._safe_sync_source, '__noretry__': True})
-            raise
-
-    def set_master_clock_rate(self, master_clock_rate):
-        """
-        Sets the master clock rate by configuring the RFDC decimation and SPLL,
-        and then resetting downstream clocks.
-        """
-        if master_clock_rate not in self.rfdc.master_to_sample_clk:
-            self.log.error('Unsupported master clock rate selection {}'
-                           .format(master_clock_rate))
-            raise RuntimeError('Unsupported master clock rate selection')
-        sample_clock_freq, decimation, is_legacy_mode, halfband = \
-                self.rfdc.master_to_sample_clk[master_clock_rate]
-        for db_idx, _ in enumerate(self.dboards):
-            db_rfdc_resamp, db_halfband = self.rfdc.get_rfdc_resampling_factor(db_idx)
-            if db_rfdc_resamp != decimation or db_halfband != halfband:
-                msg = (f'master_clock_rate {master_clock_rate} is not compatible '
-                       f'with FPGA which expected decimation {db_rfdc_resamp}')
-                self.log.error(msg)
-                raise RuntimeError(msg)
-        self.log.trace(f"Set master clock rate (SPLL) to: {master_clock_rate}")
-        self._clk_mgr.set_spll_rate(sample_clock_freq, is_legacy_mode)
-        self._master_clock_rate = master_clock_rate
-        self.rfdc.sync()
-        self._clk_mgr.config_pps_to_timekeeper(master_clock_rate)
-
     def set_trigger_io(self, direction):
         """
         Switch direction of clocking board Trigger I/O SMA socket.
@@ -998,35 +778,6 @@ class x4xx(ZynqComponents, PeriphManagerBase):
         self.log.debug("Updating mboard FPGA type info to {}".format(fpga_string))
         self.updateable_components['fpga']['type'] = fpga_string
 
-    #######################################################################
-    # Timekeeper API
-    #######################################################################
-    def get_master_clock_rate(self):
-        """ Return the master clock rate set during init """
-        return self._master_clock_rate
-
-    def get_clocks(self):
-        """
-        Gets the RFNoC-related clocks present in the FPGA design
-        """
-        # TODO: The 200 and 40 MHz clocks should not be hard coded, and ideally
-        # be linked to the FPGA image somehow
-        return [
-            {
-                'name': 'radio_clk',
-                'freq': str(self.get_master_clock_rate()),
-                'mutable': 'true'
-            },
-            {
-                'name': 'bus_clk',
-                'freq': str(200e6),
-            },
-            {
-                'name': 'ctrl_clk',
-                'freq': str(40e6),
-            }
-        ]
-
     ###########################################################################
     # GPIO API
     ###########################################################################
@@ -1272,7 +1023,7 @@ class x4xx(ZynqComponents, PeriphManagerBase):
         Return main refclock lock status. This is the lock status of the
         reference and sample PLLs.
         """
-        lock_status = self._clk_mgr.get_ref_locked()
+        lock_status = self.clk_mgr.clk_ctrl.get_ref_locked()
         return {
             'name': 'ref_locked',
             'type': 'BOOLEAN',
@@ -1298,16 +1049,6 @@ class x4xx(ZynqComponents, PeriphManagerBase):
         self.log.trace("Reading SCU internal temperature.")
         return get_temp_sensor(["EC Internal"], log=self.log)
 
-    def get_internal_temp_sensor(self):
-        """ TODO: Determine how to interpret this function """
-        self.log.warning("Reading internal temperature is not yet implemented.")
-        return {
-            'name': 'temperature',
-            'type': 'REALNUM',
-            'unit': 'C',
-            'value': '-1'
-        }
-
     def _get_fan_sensor(self, fan='fan0'):
         """ Get fan speed. """
         self.log.trace("Reading {} speed sensor.".format(fan))
@@ -1341,9 +1082,21 @@ class x4xx(ZynqComponents, PeriphManagerBase):
         assert self._gps_mgr
         self.log.trace("Reading all GPS status pins")
         return f"""
-            {self.get_gps_lock_sensor()}
+            {self.get_gps_locked_sensor()}
             {self.get_gps_alarm_sensor()}
             {self.get_gps_warmup_sensor()}
             {self.get_gps_survey_sensor()}
             {self.get_gps_phase_lock_sensor()}
         """
+
+    ###########################################################################
+    # Helpers
+    ###########################################################################
+    def pop_host_tasks(self, task):
+        """
+        Queries all known sources of host tasks and returns a list of dicts
+        that can be used for parameterization of the requested task. Depending
+        on the return value of this, the host can trigger tasks. Currently we
+        only have such tasks in the clock manager.
+        """
+        return self.clk_mgr.pop_host_tasks(task)
